searchData={"items":[{"type":"module","title":"Plumbery","doc":"Module `Plumbery` provides a DSL for building pipelines. \n\nBasic concepts are as follows:\n\n* `request` is a struct that is passed beetween pipes\n* `pipe`  is a function that accepts a struct and returns a struct, normally\nof the same kind\n* `pipeline`  is a pipe composed of other pipes. The request moves through\nthe pipeline one pipe at a time and each pipe has an opportunity to halt the\npipeline. After a pipeline is halted, the request is returned from the\npipeline without going through subsequent pipes in the pipeline\n* `inlet` is a function that accepts zero or more arguments and returns a\nrequest. Useful for API entry points that accept arbitrary arguments and\nconvert them to a request","ref":"Plumbery.html"},{"type":"module","title":"Options - Plumbery","doc":"* `:extensions` (list of module that adopts `Spark.Dsl.Extension`) - A list of DSL extensions to add to the `Spark.Dsl`\n\n* `:otp_app` (`t:atom/0`) - The otp_app to use for any application configurable options\n\n* `:fragments` (list of `t:module/0`) - Fragments to include in the `Spark.Dsl`. See the fragments guide for more.","ref":"Plumbery.html#module-options"},{"type":"module","title":"Plumbery.Request","doc":"Requests hold shared state and execution result. As the request moves through\nthe pipeline, it can be modified and later used in next steps down the\npipeline. \n\nRequests are expected to be structs with at least the following fields:\n\n* `command` is a map or struct that contains the arguments for pipeline\n* `context` contains information about calling\n  context and can be used, for example, for authorization\n* `assigns` a map with shared state\n* `result` contains pipe execution result\n* `halted?` true if pipeline is halted","ref":"Plumbery.Request.html"},{"type":"module","title":"Defining a request struct - Plumbery.Request","doc":"`Plumbery.Request` is a struct that can in many cases be used as a request for pipelines.\nIn cases when you need additional fields in your requests, you can use `Plumbery.Request`\nto define the struct.\n\nTo define a request struct, create a module and use `Plumbery.Request`,\npassing a keyword list with additional fields and their default values.\n\n    defmodule MyApp.Request do\n      @enforce_keys [:field1]\n      use Plumbery.Request, field1: nil, field2: false\n    end","ref":"Plumbery.Request.html#module-defining-a-request-struct"},{"type":"function","title":"Plumbery.Request.add_error/3","doc":"Adds an error and sets result to `{:error, errors}` where `errors`\nis a list. \n\nIf `key` is `nil`, `error` will be added to the list, otherwise `{key,\nerror}` tuple will be added.","ref":"Plumbery.Request.html#add_error/3"},{"type":"function","title":"Plumbery.Request.assign/3","doc":"Assigns a value to a key in the request's shared state map.","ref":"Plumbery.Request.html#assign/3"},{"type":"function","title":"Plumbery.Request.error/2","doc":"Sets result to `{:error, error}`.","ref":"Plumbery.Request.html#error/2"},{"type":"function","title":"Plumbery.Request.halt/1","doc":"Halts the pipeline. No further steps will be executed.","ref":"Plumbery.Request.html#halt/1"},{"type":"function","title":"Plumbery.Request.result/3","doc":"Sets result to `res`. Further steps in the pipeline will be\nexecuted and can further modify the result.\n\nWhen `strict` is true (the default), only {:error, \\_} and {:ok, \\_} tuples are\naccepted, otherwise any value can be passed.\n\nresult, not request, is the first argument intentionally. It allows for this\ntype of calls:\n\n```elixir\nsome_function(req.command)\n|> result(req)\n```","ref":"Plumbery.Request.html#result/3"},{"type":"function","title":"Plumbery.Request.success/2","doc":"Sets result to `{:ok, result}`. Further steps in the pipeline will be\nexecuted and can further modify the result.","ref":"Plumbery.Request.html#success/2"},{"type":"function","title":"Plumbery.Request.unwrap/1","doc":"Returns request's `result` field.","ref":"Plumbery.Request.html#unwrap/1"},{"type":"type","title":"Plumbery.Request.result/0","doc":"","ref":"Plumbery.Request.html#t:result/0"},{"type":"type","title":"Plumbery.Request.t/0","doc":"","ref":"Plumbery.Request.html#t:t/0"},{"type":"extras","title":"Plumbery","doc":"!file README.md\n\n\nWe recommend reading the guides in order to learn more.","ref":"intro.html"},{"type":"extras","title":"Adding Plumbery to an application","doc":"# Adding Plumbery to an application\n\nTo add Plumbery to an application, add `:plumbery` as a dependency to `mix.exs`: \n\n```elixir\ndefp deps do\n  [\n    {:plumbery, \"~> 0.1\"}\n  ]\nend\n```\n\nIf you want your code using the Plumbery DSL to be formatted as examples in this documentation, add `:plumbery` to `import_deps` in `.formatter.exs`:\n\n```elixir\n[\n  import_deps: [:plumbery],\n  inputs: [\"{mix,.formatter}.exs\", \"{config,lib,test}/**/*.{ex,exs}\"]\n]\n```","ref":"installation.html"},{"type":"extras","title":"Pipelines and pipes","doc":"# Pipelines and pipes\n<!-- vim: set tw=80 : -->\n\n\nThe main concepts of Plumbery are *requests*, *pipelines* and *pipes*.\n\nPipe is a function taking a struct called a request and returning that struct,\npotentially modifying it. We also sometimes call pipes *steps*.\nRequest is normally a `Plumbery.Request` struct.\n\nPipeline is a function taking a request struct and returning that struct,\npotentially modifying it. Note that is the exact description of a pipe, so a\npipeline is also a pipe. What makes it different is that pipelines are composed\nof pipes and call them sequentially, passing the return value of each pipe to\nthe next pipe. That looks like what Elixir's pipe operator `|>` does. But there\nis a difference: the pipeline checks the return value of each pipe and returns\nearly if there is an error or if the pipe explicitly requested pipeline \nprocessing to be stopped.\n\nThe fact that a pipeline is a pipe makes it possible to create reusable\npipelines that can be used as building blocks in other pipelines.","ref":"pipelines.html"},{"type":"extras","title":"Creating a pipeline","doc":"# Creating a pipeline\n<!-- vim: set tw=80 : -->\n\nTo create a pipeline, use `Plumbery` in a module, and use `pipeline` DSL keyword\nto declare your pipeline. \n\n!elixir examples/first_pipeline_1.exs\n\nThis example is very basic, but it demonstrates the main ideas of Plumbery.\nLet's see what the code does.\n\nFirst, the `Plumbery` module is used. This makes available the DSL. Next,\nfunctions from `Plumbery.Pipeline` are imported. This adds, among others, \n`error` and `success` functions. These functions are used to set the result of\npipe execution. `assign` function is used to set shared data that can be used in\nsubsequent pipes.\n\nNext a couple of functions is defined. They will become pipes – the building\nblocks of pipeline.\n\nFinally, a pipeline with two pipes is created. As you can see from the first\ncall example, the pipeline is compiled into a function that call the pipes in\norder and passes the request, potentially modifying it. But the second call\nexample is actually more interesting – if any pipe returns an error, the\npipeline does not call any more pipes downstream and returns immediately. This\nis the default behaviour, and it can be modified like many other aspects of\npipeline processing. Read the rest of the guides and the documentation for\n[`Plumbery.Dsl`](documentation/dsls/DSL:-Plumbery.md) module to learn more.","ref":"first_pipeline.html"},{"type":"extras","title":"Errors, escaping and halting","doc":"# Errors, escaping and halting\n<!-- vim: set tw=80 : -->\n\nAs a request is moving through a pipeline, some steps may produce errors. In\nthat case the pipeline stops and returns the first error encountered, without\ncalling the rest of the pipes.\n\nThis is not always the desired behavior. Let's consider a pipeline where the\nfirst few pipes validate different aspects of the request.\n\n!elixir examples/recovery_1.exs\n\nWhen calling the pipeline, we want it to report validation errors for all\naspects, not just the one it happens to check first. That is where the concept\nof *escaping* comes into play. By default a pipeline escapes on error, meaning\nthat as soon as the request has an error, the pipeline returns immediately\nwithout calling any further steps. After escaping is turned off, all subsequent\npipes will be called regardless of wheter the request has errors or not. That\nprovides an opportunity for counter-acting an error or adding another error that\nis orthogonal to the ones generated by previous steps.\n\nLet's modify the example so it actually does what we want.\n\n!elixir examples/recovery_2.exs\n\nSo how does `escape_on_error` options work? That's pretty simple. If it is\n`true` (the default value) then as soon as there is an error in the request, the\npipeline returns immediately. When that option is `false`, all subsequent pipes\nwill be called.\n\nPlumbery provides one more way of controlling the pipeline – pipeline halting.\nAny pipe can call `Plumbery.Request.halt/1`. That causes the pipeline to return\nimmeditely and unconditionally.\n\n\n!elixir examples/recovery_3.exs","ref":"escaping.html"},{"type":"extras","title":"Inlets, unwrapping and context","doc":"# Inlets, unwrapping and context\n<!-- vim: set tw=80 : -->","ref":"inlets.html"},{"type":"extras","title":"Inlets - Inlets, unwrapping and context","doc":"Building a request to be passed to a pipeline can be verbose, especially when\nyou just need to pass in a couple of arguments in the command field. Plumbery\nprovides a helper that makes things easier, called *inlet*. An inlet is a\ngenerated function that stuffs arguments into a request's `command` struct/map\nand calls a pipeline.\n\nLet's see how inlets work. If you read through other guides, you may have\nalready noticed that we use examples that do not look like something one would\nrealistically write but hopefully explain concepts well. I will not break this\npattern and create a pipeline that does something that does not require a\npipeline.\n\n!elixir examples/inlets_1.exs\n\nLooks like that's too much effort to just pass in a couple of numbers. Let's\ncreate and use an inlet.\n\n!elixir examples/inlets_2.exs\n\nHere we can see some interesting things about an inlet:\n\n* you can define the exact signature of the inlet function, including patterns,\n  default values and guard clauses\n* the arguments are added to request's command, under the same names as the\n  arguments\n\n\n> #### Info {: .info}\n>\n> In some cases when using named arguments and patterns it is not possible to\n> distinguish variable name from the pattern (like in `y = x` above). In that\n> case the lefthand side is used as variable name and the key to put inside the\n> command map.","ref":"inlets.html#inlets"},{"type":"extras","title":"Unwrapping - Inlets, unwrapping and context","doc":"Passing the whole request structure from pipe to pipe is necessary since we need\nto keep track of the pipeline state, but for a pipeline that is not meant to be\nincluded in other pipelines all we are interested in is the execution result,\nnot command, assigns and other stuff inside the request struct. We can make a\npipeline to return only the result using the `unwrap` option.\n\n\n!elixir examples/inlets_3.exs\n\nInlets can be also defined outside of a pipeline, in that case you need to\nspecify the pipeline to execute:\n\n!elixir examples/inlets_4.exs","ref":"inlets.html#unwrapping"},{"type":"extras","title":"Context - Inlets, unwrapping and context","doc":"Oftentimes when building APIs it is common that most entrypoints require a\ncontext argument with current user and other information. For that purpose\n`Plumbery.Request` has `context` field, but other than that Plumbery does not\nprovide support for context. The only exception is the ability to add `context`\nas the last argument for functions generated by inlets. If you specify\n`use_context` option, the `context` will be added as the last argument of the\ngenerated function.\n\n!elixir examples/inlets_5.exs","ref":"inlets.html#context"},{"type":"extras","title":"DSL: Plumbery.Dsl","doc":"<!--\nThis file was generated by Spark. Do not edit it by hand.\n-->\n# DSL: Plumbery.Dsl","ref":"dsl-plumbery.html"},{"type":"extras","title":"plumbery - DSL: Plumbery.Dsl","doc":"","ref":"dsl-plumbery.html#plumbery"},{"type":"extras","title":"Nested DSLs - DSL: Plumbery.Dsl","doc":"* [pipeline](#plumbery-pipeline)\n   * pipe\n   * escape_on_error\n   * inlet\n * [inlet](#plumbery-inlet)","ref":"dsl-plumbery.html#nested-dsls"},{"type":"extras","title":"plumbery.pipeline - DSL: Plumbery.Dsl","doc":"```elixir\npipeline name\n```","ref":"dsl-plumbery.html#plumbery-pipeline"},{"type":"extras","title":"Nested DSLs - DSL: Plumbery.Dsl","doc":"* [pipe](#plumbery-pipeline-pipe)\n * [escape_on_error](#plumbery-pipeline-escape_on_error)\n * [inlet](#plumbery-pipeline-inlet)","ref":"dsl-plumbery.html#nested-dsls"},{"type":"extras","title":"Arguments - DSL: Plumbery.Dsl","doc":"| Name | Type | Default | Docs |\n|------|------|---------|------|\n| [`name`](#plumbery-pipeline-name){: #plumbery-pipeline-name .spark-required} | `atom` |  | The name of the pipeline. This becomes the name of the generated function |","ref":"dsl-plumbery.html#arguments"},{"type":"extras","title":"Options - DSL: Plumbery.Dsl","doc":"| Name | Type | Default | Docs |\n|------|------|---------|------|\n| [`doc`](#plumbery-pipeline-doc){: #plumbery-pipeline-doc } | `String.t` |  | Documentation for generated function |\n| [`unwrap`](#plumbery-pipeline-unwrap){: #plumbery-pipeline-unwrap } | `boolean` | `false` | When true, the pipeline will unwrap the result before returning. Makes sense only for pipelines that are not meant to be used in other pipelines |\n| [`private`](#plumbery-pipeline-private){: #plumbery-pipeline-private } | `boolean` | `false` | When true, the generated pipeline function is private |","ref":"dsl-plumbery.html#options"},{"type":"extras","title":"plumbery.pipeline.pipe - DSL: Plumbery.Dsl","doc":"```elixir\npipe function\n```\n\n\nAdds a pipe to pipeline.","ref":"dsl-plumbery.html#plumbery-pipeline-pipe"},{"type":"extras","title":"Arguments - DSL: Plumbery.Dsl","doc":"| Name | Type | Default | Docs |\n|------|------|---------|------|\n| [`function`](#plumbery-pipeline-pipe-function){: #plumbery-pipeline-pipe-function } | `atom \\| {module, atom}` |  | Function to call. Can be either local function name specified as atom, or remote function specified as {Module, :function} tuple. Local functions can be private |","ref":"dsl-plumbery.html#arguments"},{"type":"extras","title":"Introspection - DSL: Plumbery.Dsl","doc":"Target: `Plumbery.Pipe`","ref":"dsl-plumbery.html#introspection"},{"type":"extras","title":"plumbery.pipeline.escape_on_error - DSL: Plumbery.Dsl","doc":"```elixir\nescape_on_error escape\n```","ref":"dsl-plumbery.html#plumbery-pipeline-escape_on_error"},{"type":"extras","title":"Arguments - DSL: Plumbery.Dsl","doc":"| Name | Type | Default | Docs |\n|------|------|---------|------|\n| [`escape`](#plumbery-pipeline-escape_on_error-escape){: #plumbery-pipeline-escape_on_error-escape } | `boolean` | `true` | When true, the pipeline will not call any more pipes as soon as one of the pipes returns an error |","ref":"dsl-plumbery.html#arguments"},{"type":"extras","title":"Introspection - DSL: Plumbery.Dsl","doc":"Target: `Plumbery.EscapeOnError`","ref":"dsl-plumbery.html#introspection"},{"type":"extras","title":"plumbery.pipeline.inlet - DSL: Plumbery.Dsl","doc":"```elixir\ninlet signature\n```","ref":"dsl-plumbery.html#plumbery-pipeline-inlet"},{"type":"extras","title":"Arguments - DSL: Plumbery.Dsl","doc":"| Name | Type | Default | Docs |\n|------|------|---------|------|\n| [`signature`](#plumbery-pipeline-inlet-signature){: #plumbery-pipeline-inlet-signature .spark-required} | `any` |  | Signature of generated function. Named arguments will be copied to request's `command` |","ref":"dsl-plumbery.html#arguments"},{"type":"extras","title":"Options - DSL: Plumbery.Dsl","doc":"| Name | Type | Default | Docs |\n|------|------|---------|------|\n| [`request`](#plumbery-pipeline-inlet-request){: #plumbery-pipeline-inlet-request } | `module` | `Plumbery.Request` | Module that provides the struct to be used as request |\n| [`command`](#plumbery-pipeline-inlet-command){: #plumbery-pipeline-inlet-command } | `module` |  | Module that provides the struct to be used as command. If not specified, a map will be used |\n| [`use_context`](#plumbery-pipeline-inlet-use_context){: #plumbery-pipeline-inlet-use_context } | `boolean` |  | When true, an aditional argument will be added to the function, and its value will be copied to request's `context` field |","ref":"dsl-plumbery.html#options"},{"type":"extras","title":"Introspection - DSL: Plumbery.Dsl","doc":"Target: `Plumbery.Inlet`","ref":"dsl-plumbery.html#introspection"},{"type":"extras","title":"Introspection - DSL: Plumbery.Dsl","doc":"Target: `Plumbery.Pipeline`","ref":"dsl-plumbery.html#introspection"},{"type":"extras","title":"plumbery.inlet - DSL: Plumbery.Dsl","doc":"```elixir\ninlet signature\n```","ref":"dsl-plumbery.html#plumbery-inlet"},{"type":"extras","title":"Arguments - DSL: Plumbery.Dsl","doc":"| Name | Type | Default | Docs |\n|------|------|---------|------|\n| [`signature`](#plumbery-inlet-signature){: #plumbery-inlet-signature .spark-required} | `any` |  | Signature of generated function. Named arguments will be copied to request's `command` |","ref":"dsl-plumbery.html#arguments"},{"type":"extras","title":"Options - DSL: Plumbery.Dsl","doc":"| Name | Type | Default | Docs |\n|------|------|---------|------|\n| [`pipeline`](#plumbery-inlet-pipeline){: #plumbery-inlet-pipeline .spark-required} | `module \\| {module, atom}` |  | Pipeline to call |\n| [`request`](#plumbery-inlet-request){: #plumbery-inlet-request } | `module` | `Plumbery.Request` | Module that provides the struct to be used as request |\n| [`command`](#plumbery-inlet-command){: #plumbery-inlet-command } | `module` |  | Module that provides the struct to be used as command. If not specified, a map will be used |\n| [`use_context`](#plumbery-inlet-use_context){: #plumbery-inlet-use_context } | `boolean` |  | When true, an aditional argument will be added to the function, and its value will be copied to request's `context` field |","ref":"dsl-plumbery.html#options"},{"type":"extras","title":"Introspection - DSL: Plumbery.Dsl","doc":"Target: `Plumbery.Inlet`\n\n\n\n\n\n .spark-required::after { content: \"*\"; color: red !important; }","ref":"dsl-plumbery.html#introspection"}],"content_type":"text/markdown","producer":{"name":"ex_doc","version":[48,46,51,52,46,50]}}